/*
  TerritoryMP.cpp - Displays territory instead of system for MP.

  Jason Hood, 29 & 30 September, 2009.
  v2.00, 28 January, 2013.

  Patches the server to show the territory a player is in, rather than the
  system, and/or to "hide" players with certain mounted equipment.  The
  territories are defined as [system] entries in DATA\UNIVERSE\universe.ini (on
  both the server and client; see TerritoryMP.ini for an example):

	[system]
	nickname = territory_name	; identify the territory
	strid_name = 1249		; string to display
	systems = nickname, ... 	; systems belonging to the territory
					; this is not necessary on the client
	msg_id_prefix = none		; needed to remove a warning

	[system]
	nickname = player_system_hider	; the "hidden" system
	strid_name = 1			; blank (a space)
	; If you want to show as "????", don't define this system on the client.
	equipment = nickname, ...	; equipment indicating a hidden player
					; this is not necessary on the client
	msg_id_prefix = none

  Caveat: when you first log in, and you're on a base, the equipment list is
	  undefined, so you will not be hidden until launch or reselection.

  Install:
	Copy to EXE and add to [Libraries] in dacomsrv.ini.
	Requires and assumes v1.1.
*/


#include "Common.h"
#include "Server.h"
#include <map>
#include <set>

#define NAKED	__declspec(naked)
#define STDCALL __stdcall


// FLPACKET_SERVER_MISCOBJUPDATE, flag = 0x84
// (called from IServerImpl::CharacterSelect & PlayerDB::SendSystemID)
#define ADDR_SYSTEM ((PDWORD)(0x6b37ed0+1))


DWORD dummy;
#define ProtectX( addr, size ) \
  VirtualProtect( (LPVOID)addr, size, PAGE_EXECUTE_READWRITE, &dummy );

#define RELOFS( from, to ) \
  *(PDWORD)(from) = (DWORD)(to) - (DWORD)(from) - 4

#define REL2ABS( addr ) ((DWORD)(addr) + *(PDWORD)(addr) + 4)

#define NEWOFS( from, to ) \
  to##_Old = REL2ABS( from ); \
  RELOFS( from, to##_Hook )


typedef std::map<UINT, UINT> SystemMap;
typedef std::set<UINT> EquipSet;

SystemMap System;
EquipSet  Equip;
UINT	  player_system_hider = CreateID( "player_system_hider" );


void STDCALL map_system( UINT& system, UINT player )
{
  if (!Equip.empty())
  {
    std::list<EquipDesc>::const_iterator iter, end;

    --player;
    iter = Players.playerdata[player].equipdesclist.equip.begin();
    end  = Players.playerdata[player].equipdesclist.equip.end();
    for (; iter != end; ++iter)
    {
      if (iter->bMounted && Equip.find( iter->iArchID ) != Equip.end())
      {
	system = player_system_hider;
	return;
      }
    }
  }

  SystemMap::const_iterator iter = System.find( system );
  if (iter != System.end())
    system = iter->second;
}


DWORD System_Old;

NAKED
void System_Hook()
{
  __asm {
	lea	eax, [esp+4]
	push	ecx
	push	[esp+4+4+4+0x10]
	push	eax
	call	map_system
	pop	ecx
	jmp	System_Old
	align	16
  }
}


// Yeah, I know, I should be using a class...
static int  range_pos, range_last;
static char range_string[128];
static int  range_nn;

// Get the first number in the range of numbers represented by S.  It is of the
// form STRINGNN-MM; anything else is treated as a nickname.  MM may be
// shortened to just the changed digits (eg: "Li01-5" is the same as "Li01-05";
// however, "Li01-Li05", "Li01-1" and "Li05-1" will all be treated as individual
// nicknames).
UINT get_first_range( LPCSTR s )
{
  range_pos = range_last = 0;

  LPCSTR n = s;
  LPCSTR d = strchr( n, '-' );
  int len = (d) ? d - n : 0;
  if (len < 3 || len >= sizeof(range_string) ||
      d[-1] < '0' || d[-1] > '9' || d[-2] < '0' || d[-2] > '9' ||
       d[1] < '0' ||  d[1] > '9' || (d[2] && (d[2] < '0' || d[2] > '9')))
  {
    return CreateID( n );
  }
  range_pos  = (d[-2] - '0') * 10 + d[-1] - '0';
  range_last = d[1] - '0';
  if (d[2])
    range_last = range_last * 10 + d[2] - '0';
  if (range_last <= range_pos)
    return CreateID( n );

  range_nn = len - 1;
  memcpy( range_string, n, len );
  range_string[len] = '\0';
  return CreateID( range_string );
}


UINT get_next_range()
{
  if (++range_pos > range_last)
    return 0;

  if (++range_string[range_nn] > '9')
  {
    range_string[range_nn] = '0';
    ++range_string[range_nn-1];
  }
  return CreateID( range_string );
}


// Read system nicknames from each parameter, storing the territory for each.
void set_territories( INI_Reader& ini, UINT territory )
{
  for (int i = ini.get_num_parameters(); --i >= 0;)
  {
    UINT id = get_first_range( ini.get_value_string( i ) );
    while (id)
    {
      System[id] = territory;
      id = get_next_range();
    }
  }
}


void Initialise()
{
  INI_Reader ini;

  // Let's just assume "data path = ..\data".
  if (!ini.open( "..\\data\\universe\\universe.ini" ))
    return;

  while (ini.read_header())
  {
    if (ini.is_header( "system" ))
    {
      UINT nickname = 0;
      while (ini.read_value())
      {
	if (ini.is_value( "nickname" ))
	  nickname = CreateID( ini.get_value_string() );
	else if (ini.is_value( "systems" ))
	{
	  if (nickname)
	    set_territories( ini, nickname );
	}
	else if (ini.is_value( "equipment" ))
	{
	  // Assume it's for player_system_hider.
	  for (int i = ini.get_num_parameters(); --i >= 0;)
	    Equip.insert( CreateID( ini.get_value_string( i ) ) );
	}
      }
    }
  }
}


void Patch( void )
{
  Initialise();

  ProtectX( ADDR_SYSTEM, 4 );
  NEWOFS( ADDR_SYSTEM, System );
}


BOOL WINAPI DllMain( HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved )
{
  if (fdwReason == DLL_PROCESS_ATTACH)
    Patch();

  return TRUE;
}
